# Generated by Django 4.2 on 2023-09-06 16:02

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_drf_otp', '0003_remove_verificationtoken_expire_date'),
    ]

    operations = [
        migrations.AlterField(
            model_name='usernotificationpreference',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='verificationtoken',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
    ]



# Generated by Django 4.2 on 2023-08-28 06:46

from django.apps.registry import Apps
from django.contrib.auth import get_user_model
from django.db import migrations


def populate_notification_preference(apps: Apps, schema_editor):
    UserNotificationPreference = apps.get_model("django_drf_otp", "UserNotificationPreference")
    User = get_user_model()

    users = User.objects.values_list("pk", flat=True)
    for user_id in users:
        UserNotificationPreference.objects.get_or_create(user_id=user_id, notification_type="email")


class Migration(migrations.Migration):
    dependencies = [
        ("django_drf_otp", "0001_initial"),
    ]

    operations = [
        migrations.RunPython(code=populate_notification_preference, reverse_code=migrations.RunPython.noop),
    ]



from typing import Optional, Union
from rest_framework import exceptions
from rest_framework import status
from django.utils.translation import gettext_lazy as _
from django.utils.encoding import force_str


class RequestNewTokenException(exceptions.APIException):
    status_code = status.HTTP_500_INTERNAL_SERVER_ERROR
    default_detail = _("You can request a new token in {seconds} seconds.")
    default_code = "request_new_token"

    def __init__(self, seconds: Union[str, int], detail: Optional[str] = None, code: Optional[str] = None):
        if detail is None:
            detail = force_str(self.default_detail).format(seconds=seconds)
        super().__init__(detail, code)



from django.db.models import TextChoices


class TokenType(TextChoices):
    TOTP = "totp"
    HOTP = "hotp"


class NotificationType(TextChoices):
    EMAIL = "email"
    SMS = "sms"
    AUTHENTICATOR = "authenticator"



from django.conf import settings
from django.db.models.signals import post_save
from django.dispatch import receiver

from .models import UserNotificationPreference
from .settings import otp_settings


@receiver(post_save, sender=settings.AUTH_USER_MODEL)
def user_is_created_or_save(sender, instance, created, **kwargs):
    notification_type = otp_settings.DEFAULT_NOTIFICATION_TYPE

    if created:
        UserNotificationPreference.objects.create(user=instance, notification_type=notification_type)



# Generated by Django 4.2 on 2023-09-04 08:25

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("django_drf_otp", "0002_populate_notification_preference"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="verificationtoken",
            name="expire_date",
        ),
    ]



import os
from pathlib import Path

from django.utils.translation import gettext_lazy as _
from environ import Env

BASE_DIR = Path(__file__).resolve().parent.parent
DOTENV_FILE_NAME = os.environ.get("DOTENV_FILE_NAME", ".env")

env = Env()
env.read_env(os.path.join(BASE_DIR, DOTENV_FILE_NAME))

SECRET_KEY = 'django-insecure-*6qn_(%hgzm8rg)ro-)rs3tvnkzwvl15%9+kh+45vvn3az6uo5'

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'django_extensions',
    'rest_framework',
    'rest_framework.authtoken',
    'django_drf_otp',
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'django_drf_otp.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

WSGI_APPLICATION = 'core.wsgi.application'

# Database
# https://docs.djangoproject.com/en/4.2/ref/settings/#databases

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.postgresql",
        "HOST": env("DB_HOST"),
        "USER": env("DB_USER"),
        "PASSWORD": env("DB_PASSWORD"),
        "NAME": env("DB_NAME"),
        "PORT": env("DB_PORT", cast=int, default=5432),
    }
}

LANGUAGES = [
    ("tr", _("Turkish")),
    ("en", _("English")),
]

LOCALE_PATHS = [
    BASE_DIR / "django_drf_otp" / "locale",
]

STATIC_URL = 'static/'

DEFAULT_AUTO_FIELD = 'django.db.models.BigAutoField'


EMAIL_HOST = env("EMAIL_HOST", cast=str, default=None)
EMAIL_HOST_USER = env("EMAIL_HOST_USER", cast=str, default=None)
EMAIL_HOST_PASSWORD = env("EMAIL_HOST_PASSWORD", cast=str, default=None)
EMAIL_PORT = env("EMAIL_PORT", cast=int, default=None)
EMAIL_EMAIL = env("EMAIL_EMAIL", cast=str, default=None)
EMAIL_EMAIL_PREFIX = env("EMAIL_EMAIL_PREFIX", cast=str, default=None)
EMAIL_USE_TLS = env("EMAIL_USE_TLS", cast=bool, default=True)
DEFAULT_FROM_EMAIL = EMAIL_EMAIL



from typing import Dict

from django.contrib.auth import get_user_model
from django.utils.translation import gettext_lazy as _
from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from .models import VerificationToken


class OTPTokenSerializer(serializers.Serializer):
    username_field = get_user_model().USERNAME_FIELD
    password = serializers.CharField(write_only=True)

    default_error_messages = {
        "no_active_account": _("No active account found with the given credentials."),
        "invalid_credentials": _(f"{username_field} or password is incorrect."),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.fields[self.username_field] = serializers.CharField()

    def validate(self, attrs: Dict) -> Dict:
        UserModel = get_user_model()
        filters = {
            self.username_field: attrs[self.username_field],
            "is_active": True,
        }

        try:
            user = UserModel.objects.get(**filters)
        except UserModel.DoesNotExist:
            raise ValidationError(
                detail=self.error_messages["no_active_account"],
                code="no_active_account",
            )

        if not user.check_password(attrs["password"]):
            raise ValidationError(
                detail=self.error_messages["invalid_credentials"].format(self.username_field),
                code="invalid_credentials",
            )

        attrs["user"] = user
        return attrs


class OTPTokenVerifySerializer(serializers.Serializer):
    token = serializers.CharField(max_length=6, min_length=6)
    verification_token = serializers.PrimaryKeyRelatedField(read_only=True)
    user = serializers.PrimaryKeyRelatedField(read_only=True)

    def validate(self, attrs: Dict) -> Dict:
        try:
            verification_token = VerificationToken.objects.get(token=attrs["token"], is_active=True)
        except VerificationToken.DoesNotExist:
            raise ValidationError(detail=_("Invalid token."), code="invalid_token")
        except VerificationToken.MultipleObjectsReturned:
            VerificationToken.objects.filter(token=attrs["token"], is_active=True).update(is_active=False)
            raise ValidationError(detail=_("Invalid token."), code="invalid_token")

        from .services import OTPService

        verified = OTPService().verify_verification_token(verification_token=verification_token)
        if not verified:
            raise ValidationError(detail=_("Token has expired."), code="token_expired")

        verification_token.is_active = False
        verification_token.save(update_fields=["is_active"])

        attrs["verification_token"] = verification_token
        attrs["user"] = verification_token.user
        return attrs


class OTPTokenResendSerializer(serializers.Serializer):
    username_field = get_user_model().USERNAME_FIELD

    default_error_messages = {
        "no_active_account": _("No active account found with the given credentials."),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.fields[self.username_field] = serializers.CharField()

    def validate(self, attrs: Dict) -> Dict:
        UserModel = get_user_model()
        filters = {
            self.username_field: attrs[self.username_field],
            "is_active": True,
        }
        try:
            user = UserModel.objects.get(**filters)
        except UserModel.DoesNotExist:
            raise ValidationError(
                detail=self.error_messages["no_active_account"],
                code="no_active_account",
            )

        attrs["user"] = user
        return attrs



# Generated by Django 4.2 on 2024-01-19 05:42

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_drf_otp', '0006_verificationtoken_active_token_idx'),
    ]

    operations = [
        migrations.AlterField(
            model_name='verificationtoken',
            name='secret_chars',
            field=models.CharField(max_length=256, null=True, verbose_name='Extra secret characters'),
        ),
    ]



from typing import Dict, Optional, Type

from django.utils.translation import gettext_lazy as _
from rest_framework import serializers, status
from rest_framework.authtoken.models import Token
from rest_framework.permissions import AllowAny
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.viewsets import GenericViewSet

from django_drf_otp.models import VerificationToken

from .enums import NotificationType
from .serializers import OTPTokenResendSerializer, OTPTokenSerializer, OTPTokenVerifySerializer
from .services import OTPService
from .settings import otp_settings


class OTPTokenViewSet(GenericViewSet):
    permission_classes = (AllowAny,)
    service = OTPService()
    LOGIN = "login"
    VERIFY_TOKEN = "verify_token"
    RESEND_TOKEN = "resend_token"
    serializer_classes = {
        LOGIN: OTPTokenSerializer,
        VERIFY_TOKEN: OTPTokenVerifySerializer,
        RESEND_TOKEN: OTPTokenResendSerializer,
    }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.OTP_ENABLED = otp_settings.ENABLE_OTP_AUTHENTICATION

    def login(self, request: Request, *args, **kwargs) -> Response:
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        verification_token: Optional[VerificationToken] = None
        if self.OTP_ENABLED:
            extra_context = self.get_extra_context_data(request=request, validated_data=serializer.validated_data)
            verification_token = self.service.generate_and_notify_token(
                user=serializer.validated_data["user"], extra_context=extra_context
            )

        data = self.get_success_response_data(
            verification_token=verification_token, user=serializer.validated_data["user"]
        )
        return Response(data=data, status=status.HTTP_200_OK)

    def verify_token(self, request: Request, *args, **kwargs) -> Response:
        if not self.OTP_ENABLED:
            return Response(data={"detail": _("Otp authentication is disabled.")}, status=status.HTTP_400_BAD_REQUEST)

        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        data = self.get_success_response_data(**serializer.validated_data)
        return Response(data=data, status=status.HTTP_200_OK)

    def resend_token(self, request: Request, *args, **kwargs) -> Response:
        if not self.OTP_ENABLED:
            return Response(data={"detail": _("Otp authentication is disabled.")}, status=status.HTTP_400_BAD_REQUEST)

        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        extra_context = self.get_extra_context_data(request=request, validated_data=serializer.validated_data)
        verification_token = self.service.resend_token(
            user=serializer.validated_data["user"], extra_context=extra_context
        )

        data = self.get_success_response_data(verification_token=verification_token)
        return Response(data=data, status=status.HTTP_200_OK)

    def get_serializer_class(self, *args, **kwargs) -> Type[serializers.Serializer]:
        return self.serializer_classes[self.action]

    def get_success_response_data(self, **kwargs) -> Dict:
        if self.action in [self.LOGIN, self.RESEND_TOKEN] and self.OTP_ENABLED:
            verification_token: VerificationToken = kwargs["verification_token"]
            return {
                "detail": _("OTP token sent successfully."),
                "otp_ttl": otp_settings.TOTP_TOKEN_TTL,
                "otp_created_date": verification_token.created_date.isoformat(),
            }
        else:
            token, created = Token.objects.get_or_create(user=kwargs["user"])
            return {"token": token.key}

    def get_extra_context_data(self, request: Request, validated_data: Dict) -> Dict:
        context = {"request": request}

        user = validated_data["user"]
        default_notification_type = otp_settings.DEFAULT_NOTIFICATION_TYPE

        nofification_type = (
            user.notification_preference.notification_type
            if hasattr(user, "notification_preference")
            else default_notification_type
        )
        if nofification_type == NotificationType.EMAIL:
            context.update(
                {"first_name": user.first_name, "last_name": user.last_name, "full_name": user.get_full_name()}
            )

        return context



from typing import Dict, Optional, Type

from django.contrib.auth.models import AbstractUser
from django.utils import timezone

from django_drf_otp.exceptions import RequestNewTokenException

from .enums import NotificationType, TokenType
from .models import VerificationToken
from .notifiers import BaseNotifier, EmailNotifier
from .settings import otp_settings
from .strategies import BaseOTPStrategy, TOTPStrategy

TOKEN_STRATEGY_MAP = {
    TokenType.TOTP: TOTPStrategy,
}
TOKEN_NOTIFIER_MAP = {
    NotificationType.EMAIL: EmailNotifier,
}


class OTPService:
    def verify_token(
        self, token: str, token_type: Optional[TokenType] = None, extra_secret_chars: Optional[str] = None
    ) -> bool:
        strategy_class = self._get_strategy_class(token_type=token_type)
        return strategy_class.verify_token(token=token, extra_secret_chars=extra_secret_chars)

    def verify_verification_token(self, verification_token: VerificationToken) -> bool:
        if (timezone.now() - verification_token.created_date).seconds > otp_settings.TOTP_TOKEN_TTL:
            return False

        return True

    def generate_token(self, token_type: Optional[TokenType] = None, extra_secret_chars: Optional[str] = None) -> str:
        strategy_class = self._get_strategy_class(token_type=token_type)
        return strategy_class.generate_token(extra_secret_chars=extra_secret_chars)

    def notify_user(self, user: Type[AbstractUser], token: str, extra_context: Optional[Dict] = None) -> None:
        notification_type = otp_settings.DEFAULT_NOTIFICATION_TYPE
        if hasattr(user, "notification_preference"):
            notification_type = user.notification_preference.notification_type

        notifier_class = self._get_notifier_class(notification_type=notification_type)
        notifier_class.notify(email=user.email, token=token, extra_context=extra_context or {})

    def generate_and_notify_token(
        self, user: Type[AbstractUser], token_type: Optional[TokenType] = None, extra_context: Optional[Dict] = None
    ) -> VerificationToken:
        strategy_class = self._get_strategy_class(token_type=token_type)
        verification_token = strategy_class.generate_token_for_user(user=user)
        self.notify_user(user=user, token=verification_token.token, extra_context=extra_context or {})
        return verification_token

    def resend_token(self, user: Type[AbstractUser], extra_context: Optional[Dict] = None) -> VerificationToken:
        last_active_token = (
            VerificationToken.objects.filter(user=user, is_active=True).order_by("-created_date").first()
        )

        if (
            last_active_token
            and (timezone.now() - last_active_token.created_date).seconds < otp_settings.TOTP_TOKEN_TTL
        ):
            raise RequestNewTokenException(seconds=otp_settings.TOTP_TOKEN_TTL)

        return self.generate_and_notify_token(user=user, extra_context=extra_context)

    def _get_notifier_class(self, notification_type: NotificationType) -> BaseNotifier:
        if notification_type not in TOKEN_NOTIFIER_MAP:
            raise ValueError(
                f"'{notification_type}' is not a valid notification type. Valid notification types are '{', '.join(TOKEN_NOTIFIER_MAP.keys())}'"
            )

        notifier_class = TOKEN_NOTIFIER_MAP[notification_type]
        return notifier_class()

    def _get_strategy_class(self, token_type: Optional[TokenType] = None) -> BaseOTPStrategy:
        if not token_type:
            token_type = otp_settings.DEFAULT_TOKEN_TYPE

        if token_type not in TOKEN_STRATEGY_MAP:
            raise ValueError(
                f"'{token_type}' is not a valid token type. Valid token types are '{', '.join(TOKEN_STRATEGY_MAP.keys())}'"
            )

        strategy_class = TOKEN_STRATEGY_MAP.get(token_type)
        return strategy_class()



# Generated by Django 4.2 on 2024-01-09 08:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_drf_otp', '0005_verificationtoken_secret_chars_and_more'),
    ]

    operations = [
        migrations.AddIndex(
            model_name='verificationtoken',
            index=models.Index(fields=['is_active', 'token'], name='active_token_idx'),
        ),
    ]



from django.urls import path

from .viewsets import OTPTokenViewSet

urlpatterns = [
    path("login/", OTPTokenViewSet.as_view({"post": "login"}), name="login"),
    path("verify-token/", OTPTokenViewSet.as_view({"post": "verify_token"}), name="verify-token"),
    path("resend-token/", OTPTokenViewSet.as_view({"post": "resend_token"}), name="resend-token"),
]
